
/*
 * $Id: histograms.cc,v 1.1.1.3 1999/03/08 07:56:07 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "ILog.h"
#include "OLog.h"

#include "histograms.h"



/* Log2Hist */

/*
 *   pow2       min :        max /precision =#entries (accum size)
 *     8          0 :        511 /        1 = 512  (  512 or   2K)
 *     9        512 :       1023 /        2 = 256  (  768 or   3K)
 *    10       1024 :       2047 /        4 = 256  ( 1024 or   4K)
 *    11       2048 :       4095 /        8 = 256  ( 1280 or   5K)
 *    12       4096 :       8191 /       16 = 256  ( 1536 or   6K)
 *    13       8192 :      16383 /       32 = 256  ( 1792 or   7K)
 *    14      16384 :      32767 /       64 = 256  ( 2048 or   8K)
 *    15      32768 :      65535 /      128 = 256  ( 2304 or   9K)
 *    16      65536 :     131071 /      256 = 256  ( 2560 or  10K)
 *    17     131072 :     262143 /      512 = 256  ( 2816 or  11K)
 *    18     262144 :     524287 /     1024 = 256  ( 3072 or  12K)
 *    19     524288 :    1048575 /     2048 = 256  ( 3328 or  13K)
 *    20    1048576 :    2097151 /     4096 = 256  ( 3584 or  14K)
 *    21    2097152 :    4194303 /     8192 = 256  ( 3840 or  15K)
 *    22    4194304 :    8388607 /    16384 = 256  ( 4096 or  16K)
 *    23    8388608 :   16777215 /    32768 = 256  ( 4352 or  17K)
 *    24   16777216 :   33554431 /    65536 = 256  ( 4608 or  18K)
 *    25   33554432 :   67108863 /   131072 = 256  ( 4864 or  19K)
 *    26   67108864 :  134217727 /   262144 = 256  ( 5120 or  20K)
 *    27  134217728 :  268435455 /   524288 = 256  ( 5376 or  21K)
 *    28  268435456 :  536870911 /  1048576 = 256  ( 5632 or  22K)
 *    29  536870912 : 1073741823 /  2097152 = 256  ( 5888 or  23K)
 *    30 1073741824 : 2147483647 /  4194304 = 256  ( 6144 or  24K)
 */

// XXX: change 30 to 31(-1) and replace the meaning of theMax!!

// note: val2Bin has static binding here
Log2Hist::Log2Hist(): Histogram(0, 1<<30, val2Bin(1<<30)) {
}

// note: val2Bin has static binding here
Log2Hist::Log2Hist(Val aMin, Val aMax): 
	Histogram(aMin, aMax, val2Bin(aMax-aMin)) {
}

int Log2Hist::val2Bin(Val v) const {
	Assert(v >= 0);
	const int vo = v; 
	if (v >>= 9) {
		int p = 1;
		while (v >>= 1) p++; // count powers of 2; (super-bin id)
		const int min_val = 1 << (8+p);
		const int sbin_offset = (vo - min_val) >> p;
		const int sbin = (p+1) << 8;
		return sbin + sbin_offset;
	} else
		return vo;
}

Log2Hist::Val Log2Hist::bin2Val(int bin) const {
	const int bino = bin;
	if (bin >= 512) {
		const int bin_offset = bin % 256;
		const int sbin = bin - bin_offset;
		const int p = sbin/256 - 1;
		const int val_offset = bin_offset << p;
		const int min_val = 1 << (8+p);
		return min_val + val_offset;
	} else
		return bino;
}


/* TimeHist */

TimeHist::TimeHist(): theTmMax(Time::Msec(theValMax)) {
	Assert(theTmMax > 0);
}

/* TimeVsSizeHist */

TimeVsSizeHist::TimeVsSizeHist(): theMaxSize(Size::KB(128)-1) {
	const int count = 1+szGroup(theMaxSize);
	theHists.stretch(count);
	for (int i = 0; i < count; ++i)
		theHists.append(new TimeHist);
}

TimeVsSizeHist::~TimeVsSizeHist() {
	while (theHists.count()) delete theHists.pop();
}

static inline
OLog &operator <<(OLog &log, const TimeHist *h) {
	Assert(h);
	return log << *h;
}

void TimeVsSizeHist::store(OLog &log) const {
	log << theMaxSize << theHists;
}

static inline
ILog &operator >>(ILog &log, TimeHist *h) {
	Assert(h);
	return log >> *h;
}

void TimeVsSizeHist::load(ILog &log) {
	const int ms = log.geti();
	Must(ms <= theMaxSize);
	theMaxSize = ms;
	log >> theHists;
}

// log2 step except a special case of 100 bytes
// range must be checked in the caller
int TimeVsSizeHist::szGroup(Size sz) const {
	if (sz < 100)
		return 0;
	int s = sz /= 1024;  // KB
	int n = 1;
	while (s >>= 1) n++; // log2
	return n;
}

void TimeVsSizeHist::add(const TimeVsSizeHist &h) {
	// histograms must match
	Assert(theMaxSize == h.theMaxSize);
	Assert(theHists.count() == h.theHists.count());

	for (int i = 0; i < theHists.count(); ++i)
		theHists[i]->add(*h.theHists[i]);
}
